/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { ChangeAtomId, ChangesetLocalId, RevisionMetadataSource, RevisionTag } from "../../core/index.js";
import { RangeMap } from "../../util/index.js";
import { CrossFieldManager, CrossFieldQuerySet } from "../modular-schema/index.js";
import { Attach, Detach, Insert, LineageEvent, Mark, NoopMark, Changeset, Remove, CellId, CellMark, AttachAndDetach, MarkEffect } from "./types.js";
import { EmptyInputCellMark, MoveMarkEffect, DetachOfRemovedNodes, CellRename } from "./helperTypes.js";
export declare function isEmpty<T>(change: Changeset<T>): boolean;
export declare function createEmpty<T>(): Changeset<T>;
export declare function isNewAttach(mark: Mark<unknown>, revision?: RevisionTag): boolean;
export declare function isNewAttachEffect(effect: MarkEffect, cellId: CellId | undefined, revision?: RevisionTag): boolean;
export declare function isInsert(mark: MarkEffect): mark is Insert;
export declare function isAttach(effect: MarkEffect): effect is Attach;
export declare function isReattach(mark: Mark<unknown>): boolean;
export declare function isReattachEffect(effect: MarkEffect, cellId: CellId | undefined): boolean;
export declare function isActiveReattach<T>(mark: Mark<T>): mark is CellMark<Insert, T> & {
    conflictsWith?: undefined;
};
export declare function areEqualCellIds(a: CellId | undefined, b: CellId | undefined): boolean;
export declare function getInputCellId(mark: Mark<unknown>, revision: RevisionTag | undefined, metadata: RevisionMetadataSource | undefined): CellId | undefined;
export declare function getOutputCellId(mark: Mark<unknown>, revision: RevisionTag | undefined, metadata: RevisionMetadataSource | undefined): CellId | undefined;
export declare function cellSourcesFromMarks(marks: readonly Mark<unknown>[], revision: RevisionTag | undefined, metadata: RevisionMetadataSource | undefined, contextGetter: typeof getInputCellId | typeof getOutputCellId): Set<RevisionTag | undefined>;
export declare enum CellOrder {
    SameCell = 0,
    OldThenNew = 1,
    NewThenOld = 2
}
/**
 * Determines the order of two cells from two changesets.
 *
 * This function makes the following assumptions:
 * 1. The cells represent the same context.
 * 2. `oldMarkCell` is from a mark in a changeset that is older than the changeset that contains the mark that
 * `newMarkCell` is from.
 * 3. In terms of sequence index, all cells located before A are also located before B,
 * and all cells located before B are also located before A.
 * 4. If a changeset has a mark/tombstone that describes a cell named in some revision R,
 * then that changeset must contain marks/tombstones for all cells named in R as well as all cells named in later
 * revisions up to its own.
 * 5. If a changeset foo is rebased over a changeset bar, then the rebased version of foo must contain tombstones or
 * marks for all cells referenced or named in bar. It has yet to be determined whether this assumption is necessary
 * for the logic below.
 *
 * @param oldMarkCell - The cell referenced or named by a mark or tombstone from the older changeset.
 * @param newMarkCell - The cell referenced or named by a mark or tombstone from the newer changeset.
 * @param oldChangeKnowledge - The set of revisions that the older changeset has cell representations for.
 * @param newChangeKnowledge - The set of revisions that the newer changeset has cell representations for.
 * @param metadata - Revision metadata for the operation being carried out.
 * @returns a {@link CellOrder} which describes how the cells are ordered relative to one-another.
 */
export declare function compareCellPositionsUsingTombstones(oldMarkCell: ChangeAtomId, newMarkCell: ChangeAtomId, oldChangeKnowledge: ReadonlySet<RevisionTag | undefined>, newChangeKnowledge: ReadonlySet<RevisionTag | undefined>, metadata: RevisionMetadataSource): CellOrder;
export declare function getDetachOutputId(mark: Detach, revision: RevisionTag | undefined, metadata: RevisionMetadataSource | undefined): ChangeAtomId;
/**
 * Preserves the semantics of the given `mark` but repackages it into a `DetachOfRemovedNodes` when possible.
 */
export declare function normalizeCellRename<TNodeChange>(mark: CellMark<AttachAndDetach, TNodeChange>): CellMark<AttachAndDetach | DetachOfRemovedNodes, TNodeChange>;
/**
 * Preserves the semantics of the given `mark` but repackages it into an `AttachAndDetach` mark if it is not already one.
 */
export declare function asAttachAndDetach<TNodeChange>(mark: CellMark<CellRename, TNodeChange>): CellMark<AttachAndDetach, TNodeChange>;
export declare function cloneMark<TMark extends Mark<TNodeChange>, TNodeChange>(mark: TMark): TMark;
export declare function cloneMarkEffect<TEffect extends MarkEffect>(effect: TEffect): TEffect;
export declare function cloneCellId(id: CellId): CellId;
/**
 * @param mark - The mark to get the length of.
 * @param ignorePairing - When true, the length of a paired mark (e.g. MoveIn/MoveOut) whose matching mark is not active
 * will be treated the same as if the matching mark were active.
 * @returns The number of nodes within the output context of the mark.
 */
export declare function getOutputLength(mark: Mark<unknown>, ignorePairing?: boolean): number;
/**
 * @param mark - The mark to get the length of.
 * @returns The number of nodes within the input context of the mark.
 */
export declare function getInputLength(mark: Mark<unknown>): number;
export declare function markEmptiesCells(mark: Mark<unknown>): boolean;
export declare function markFillsCells(mark: Mark<unknown>): boolean;
export declare function markHasCellEffect(mark: Mark<unknown>): boolean;
export declare function isAttachAndDetachEffect(effect: MarkEffect): effect is AttachAndDetach;
export declare function isDetachOfRemovedNodes(mark: Mark<unknown>): mark is CellMark<DetachOfRemovedNodes, unknown>;
export declare function getDetachIdForLineage(mark: MarkEffect, fallbackRevision: RevisionTag | undefined): ChangeAtomId | undefined;
export declare function isImpactfulCellRename(mark: Mark<unknown>, revision: RevisionTag | undefined, revisionMetadata: RevisionMetadataSource): mark is CellMark<CellRename, unknown>;
export declare function areInputCellsEmpty<T>(mark: Mark<T>): mark is EmptyInputCellMark<T>;
export declare function areOutputCellsEmpty(mark: Mark<unknown>): boolean;
/**
 * Creates a mark that is equivalent to the given `mark` but with effects removed if those have no impact in the input
 * context of that mark.
 *
 * @param mark - The mark to settle. Never mutated.
 * @param revision - The revision associated with the mark.
 * @param revisionMetadata - Metadata source for the revision associated with the mark.
 * @returns either the original mark or a shallow clone of it with effects stripped out.
 */
export declare function settleMark<TChildChange>(mark: Mark<TChildChange>, revision: RevisionTag | undefined, revisionMetadata: RevisionMetadataSource): Mark<TChildChange>;
/**
 * @returns true, iff the given `mark` would have impact on the field when applied.
 * Ignores the impact of nested changes.
 * CellRename effects are considered impactful if they actually change the ID of the cells.
 */
export declare function isImpactful(mark: Mark<unknown>, revision: RevisionTag | undefined, revisionMetadata: RevisionMetadataSource): boolean;
export declare function isTombstone<T>(mark: Mark<T>): mark is CellMark<NoopMark, T> & {
    cellId: CellId;
};
export declare function isNoopMark<T>(mark: Mark<T>): mark is CellMark<NoopMark, T>;
/**
 * @returns The number of cells in the range which come before the position described by `lineage`.
 */
export declare function getOffsetInCellRange(lineage: LineageEvent[] | undefined, revision: RevisionTag | undefined, id: ChangesetLocalId, count: number): number | undefined;
export declare function areOverlappingIdRanges(id1: ChangesetLocalId, count1: number, id2: ChangesetLocalId, count2: number): boolean;
export declare function compareCellsFromSameRevision(cell1: CellId, count1: number, cell2: CellId, count2: number): number | undefined;
export declare function isDetach(mark: MarkEffect | undefined): mark is Detach;
export declare function isRemoveMark<TNodeChange>(mark: Mark<TNodeChange> | undefined): mark is CellMark<Remove, TNodeChange>;
/**
 * Attempts to extend `lhs` to include the effects of `rhs`.
 * @param lhs - The mark to extend.
 * @param rhs - The effect so extend `rhs` with.
 * @returns `lhs` iff the function was able to mutate `lhs` to include the effects of `rhs`.
 * When `undefined` is returned, `lhs` is left untouched.
 */
export declare function tryMergeMarks<T>(lhs: Mark<T>, rhs: Readonly<Mark<T>>): Mark<T> | undefined;
/**
 * @internal
 */
export interface CrossFieldTable<T = unknown> extends CrossFieldManager<T> {
    srcQueries: CrossFieldQuerySet;
    dstQueries: CrossFieldQuerySet;
    isInvalidated: boolean;
    mapSrc: Map<RevisionTag | undefined, RangeMap<T>>;
    mapDst: Map<RevisionTag | undefined, RangeMap<T>>;
    reset: () => void;
}
/**
 * @internal
 */
export declare function newCrossFieldTable<T = unknown>(): CrossFieldTable<T>;
/**
 * Splits the `mark` into two marks such that the first returned mark has length `length`.
 * @param mark - The mark to split.
 * @param revision - The revision of the changeset the mark is part of.
 * @param length - The desired length for the first of the two returned marks.
 * @param genId - An ID allocator
 * @param moveEffects - The table in which to record splitting of move marks
 * @param recordMoveEffect - Whether when splitting a move an entry should be added to `moveEffects` indicating that the mark should be split (in case we process this mark again).
 * An entry is always added to `moveEffects` indicating that the opposite end of the move should be split.
 * @returns A pair of marks equivalent to the original `mark`
 * such that the first returned mark has input length `length`.
 */
export declare function splitMark<T, TMark extends Mark<T>>(mark: TMark, length: number): [TMark, TMark];
export declare function splitMarkEffect<TEffect extends MarkEffect>(effect: TEffect, length: number): [TEffect, TEffect];
/**
 * @returns -1 if the lineage indicates that cell1 is earlier in the field than cell2.
 * Returns 1 if cell2 is earlier in the field.
 * Returns 0 if the order cannot be determined from the lineage.
 */
export declare function compareLineages(cell1: CellId, cell2: CellId): number;
export declare function extractMarkEffect<TEffect extends MarkEffect>(mark: CellMark<TEffect, unknown>): TEffect;
export declare function omitMarkEffect<TChildChange>(mark: CellMark<unknown, TChildChange>): CellMark<NoopMark, TChildChange>;
export declare function withNodeChange<TMark extends CellMark<TKind, TNodeChange>, TKind extends MarkEffect, TNodeChange>(mark: TMark, changes: TNodeChange | undefined): TMark;
export declare function withRevision<TMark extends Mark<unknown>>(mark: TMark, revision: RevisionTag | undefined): TMark;
export declare function getEndpoint(effect: MoveMarkEffect, revision: RevisionTag | undefined): ChangeAtomId;
//# sourceMappingURL=utils.d.ts.map